/*
|       FILE: UNTIL.H
|
|	This module contains copyrighted source code for part the 
|	UNTIL Language Materials. 
|
|	Written by:
|		Norman E. Smith, CDP
|		Copyright 1992, 1994
|		All Rights Reserved
|
|	Right to use, copy, and modify this code is granted
|	for personal non-commercial use, provided that this
|	copyright disclosure remains on ALL copies. Any other
|	use, reproduction, or distribution is covered in the
|	License Agreement with the Until Language Maeterials
|	documentation and in the file LICENSE.TXT.
|
| DATE CHANGED:
|       10/07/92        Added "hld". [nes]
|       12/21/92        Broke function prototypes out into a
|                       separate header file.
|	03/11/93	Added FREAD_WA [nes].
|	06/28/93	Changed tib, pstack, rstack, & here_here from
|			arrays to a char ptrs. [nes]
|	06/29/93	Changed pad from array to ptrs. [nes]
|	07/05/93	Added vocabulary stack. [nes]
|	05/30/94	Added flags to control C string words. [nes]
|	07/06/94	Added new File Descriptor structure. [nes].
|	08/04/94	Made Beta Distribution compile changes. [nes].
|	08/31/94	Expanded C Stack to 32k.
|	12/30/94	Started V2.1.1
|	01/20/95	Moved file descriptor stack to io.h. [nes] 
*/
/*
|			Magic settings for Param. Stack
*/
#define ADDR             void*
#define NUMB             long
/*
|			 Buffer sizes
*/
#define MAX_NAME         32
#define TIBSIZE          256
#define PADSIZE          128
#define PSTACKSIZE       64
#define RSTACKSIZE       32
#define PFA_SIZE         100

#define NUM_FILES        8
#define BIN              0
#define SOURCE           1
#define OPENED           2
#define CLOSED           4

#define TRUE             1
#define FALSE            0
#define BLANK            ' '
#define ESC              '\033'
				/* Values for file I/O */
#define READ            "r"
#define WRITE           "w"
#define APPEND          "a"
#define BINARY          "b"
#define TEXT		"t"
#define UPDATE          "r+"
#define NEW_UPDATE      "w+"
#define APPEND_UPDATE	"a+"

#define FOREVER         for(;;)
#define WORD_FOUND      1
#define WORD_NOT_FOUND  0
#define WORD_IMMEDIATE  -1
#define NOT_IMMEDIATE   0
#define COMPILE_MODE    1
#define EXECUTION_MODE  0
#define FIRST_CELL	1
				/* Values for smudge */
#define VISIBLE		0
#define HIDDEN		1

#undef SENGINE

union cell_types{
	char  C;
	short S;
	int   I;
	long  L;
	void  *P;
};
struct NameField{
	int  len;
	int  system;
	int  immediate;
	int  smudge;
	char *name;
};

union pfa_type{
	struct DictHeader **Waddr;
	struct DictHeader *DHaddr;
	long              lvalue;
	char              cvalue;
	long              *l_ptr;
	char              *c_ptr;
};

struct DictHeader{
	void              (*CFA)();     /* ptr to primitive function */
	union  pfa_type   PFA;          /* ptr to list of words to execute */
	struct NameField  *NFA; /* ptr to name string */
	struct DictHeader *LFA; /* points to next word in dict. */
};

struct encrypt_key{
	char crypt_key[80];
	long crypt_offset;
	long crypt_len;
} ;

struct file_descriptor{
	FILE *fd;		/* File descriptor   */
	void *read_wa;		/* Read function for this file */
	long type;		/* BIN or SOURCE               */
	char *crypt;		/* Encryption key for BIN files*/
	long status;		/* OPEN or CLOSED              */
	long position;		/* Save ftell() return value   */
	char *name;		/* Filename                    */
	char *tib;		/* TIB for this file           */
	long in;		/* IN for this file            */
} ;
#ifdef VAX
long errno;
#endif

#ifdef until

#ifdef  BCC
unsigned _stklen = 32768;		/* Increased stack size in BC++ */
#endif
/********************************************************
|               CONSTANTS FOR BUFFER SIZES AND SUCH
|
**********************************************************
|               STARTUP DATA STRUCTURE
*/
 char *comma_data;
#ifdef SENGINE
 int  show_copy          = 0;
#else
 int  show_copy          = 1;
#endif
 int  u_errno            = 0;
 int  sys_errno          = 0;
 int  Vers               = 1;
 char CALC_VERSION[]     = "2.2";
#ifndef SPARC 
 char DATE[]             = __DATE__;
#else
 char DATE[]             = "January 1995";
#endif
 char boot_word[32]      = "<outer>";
 char first_word[32]     = "";
 char dict_file[80]      = "";
 char VERSION[]          = "2.0";
 char editor[]           = "emacs";
 char autoload[80]       = "applic.app";
/*  char binload[128]        = "o:\\sgml\\bin\\until.bin";  */
#ifdef SEAL_MODULE
 char binload[128]        = "until.bin";
#else
 char binload[128]        = "until.app";
#endif
 char arg_file[128]       = "";

/********************************************************
|               UNTIL REGISTERS
*/
/*
struct file_descriptor source[NUM_FILES];
long INPUT_SOURCE;      / * Where the compiler reads from * /
long this_file;
long file_eof;
long binary_input_mode;
*/
long seal_eof;
long tib_len;
long pfa_offset;
long comma_offset;
long SP;                /* Param Stack Pointer */
long RS;                /* Return Stack Pointer */
long START = 1;         /* Startup flag */
long IN;
long STATE;             /* compile state 0=Interpret 1=compile  */
long BASE;              /* Current base                         */
long SPAN;
long hld;               /* Used in <# # #> sequences            */

/***********************+------------------+
|                       | Vocabulary Stuff |
|                       +------------------+
*/
struct DictHeader *LATEST;	/* Last word compiled          */
struct DictHeader *CURRENT;	/* Vocab words compiled into   */
struct DictHeader *CONTEXT;	/* First vocab in search order */
struct DictHeader *FORTH;
struct DictHeader *VOC_LINK;	/* Last vocab created          */
struct DictHeader *vocs[8];	/* allow 8 vocs at a time      */
struct DictHeader *VOC_FLAG;	/* Set when vocab executed in compile
				|  mode to the current value of CONTEXT.
				|  Should be used to reset CONTEXT by
				|  compiler the next time a word is found.
				|  This keeps current and context in
				|  sequence.
				*/

long              vp;		/* voc stack ptr               */
/****************************************************************
|                       Buffers and other Data Structures
*/
char *tib;
char *here_here;
long *pstack;
long *rstack;
char *pad;
ADDR R_addr;

/************************************************************
|                       OTHER STUFF
*/
int I;
union cell_types cell;

long show_stack		= TRUE;		/* T=return values */
long string_return	= TRUE;		/* T=return values */
long quote_null		= FALSE;	/* F=counted T=null*/
long left_margin        = 0;
long right_margin       = 79;
long does_flag;
long ABORT_FLAG;
long QUIT;
long ABORT;
long fargc;
char **fargv;

#ifdef SENGINE
char PROMPT_STR[20]     = "S-Engine> ";
#else
char PROMPT_STR[20]     = "Until> ";
#endif

char the_cold_msg[]     = "Cold Start...\n";
char the_warm_msg[]     = "\n";
struct encrypt_key this_key;
/*************************************************************
|               VECTORED WORDS
*/
char *(*FREAD_WA)();
char *(*READ_WA)();
char *(*Semi_colon)();

#ifdef VECTORED_IO
char *(*verbatim_vect)();
char *(*display_vect)();
char *(*center_vect)();
char *(*dot_quote_vect)();
char *(*type_vect)();
char *(*dot_vect)();
char *(*prints_vect)();
char *(*emit_vect)();
char *(*space_vect)();
char *(*spaces_vect)();
char *(*key_vect)();
char *(*query_terminal_vect)();
char *(*cr_vect)();
#endif
/*
char *(*)();
char *(*)();
char *(*)();
*/
/*************************************************************
|
*/
struct DictHeader *pfa_list[PFA_SIZE];  /* Temp area to compile list of pfa's */
struct DictHeader **IP;         	/* Instrcution Pointer */
struct DictHeader *DP;                  /* Dictionary Pointer */
struct DictHeader *WA;                  /* Word Address Register */
struct DictHeader *LIT_WA;              /* lit Word Address */
struct DictHeader *ZERO_BRAN_WA;
struct DictHeader *BRANCH_WA;
struct DictHeader *QUOTE_WA;
struct DictHeader *DOT_QUOTE_WA;
struct DictHeader *LOOP_WA;
struct DictHeader *DO_WA;
struct DictHeader *I_WA;
struct DictHeader *NAME0_WA;
struct DictHeader *ABORT_QUOTE_WA;
struct DictHeader *PLUS_LOOP_WA;
struct DictHeader *J_WA;
struct DictHeader *K_WA;
struct DictHeader *TICK_WA;
struct DictHeader *DO_COMPILE_WA;
struct DictHeader *DO_DOES_WA;
struct DictHeader *COMPILE_DO_DOES_WA;
struct DictHeader *DO_VOC_WA;
struct DictHeader *DO_DEF_WA;
struct DictHeader *DP0;
struct DictHeader *FORTH_WA;

struct NameField  *NAME0;

#else

/********************************************************
|               UNTIL REGISTERS
*/
/*
extern struct file_descriptor source[];
extern long INPUT_SOURCE;       / * Where the compiler reads from * /
extern long this_file;
extern long file_eof;
extern long binary_input_mode;
*/
extern long seal_eof;
extern long tib_len;
extern long pfa_offset;
extern long comma_offset;
extern long SP;         /* Param Stack Pointer */
extern long RS;                /* Return Stack Pointer */
extern long START;              /* Startup flag */
extern long IN;
extern long STATE;              /* compile state 0=Interpret 1=compile */
extern long INPUT_SOURCE;       /* Where the compiler reads from */
extern long BASE;               /* Current base                        */
extern long SPAN;
extern long hld;               /* Used in <# # #> sequences            */

/**********************************************************
|               STARTUP DATA STRUCTURE
*/
extern char *comma_data;
extern int  u_errno;
extern int  show_copy;
extern int  sys_errno;
extern char CALC_VERSION[];
extern int  Vers;
extern char DATE[];
extern char boot_word[];
extern char first_word[];
extern char dict_file[];
extern char VERSION[];
extern char editor[];
extern char autoload[];
extern char binload[];
extern char arg_file[];

/***********************+------------------+
|                       | Vocabulary Stuff |
|                       +------------------+
*/
extern struct DictHeader *LATEST;
extern struct DictHeader *CURRENT;
extern struct DictHeader *CONTEXT;
extern struct DictHeader *FORTH;
extern struct DictHeader *VOC_LINK;
extern struct DictHeader *vocs[8];
extern struct DictHeader *VOC_FLAG;
extern long              vp;

/****************************************************************
|                       Buffers and other Data Structures
*/
/*	NOTE: These have been changed to pointers...
extern char tib[];
extern char here_here[];
extern long pstack[];
extern long rstack[];
extern char pad[];
*/
extern char *tib;
extern char *here_here;
extern long *pstack;
extern long *rstack;
extern char *pad;
extern ADDR R_addr;

/************************************************************
|                       OTHER STUFF
*/
extern long show_stack;
extern long string_return;
extern long quote_null;

extern long left_margin;
extern long right_margin;

extern long does_flag;
extern long ABORT_FLAG;
extern long QUIT;
extern long ABORT;

/* extern static long QUIT; */
/* extern static long ABORT; */
extern long fargc;
extern char **fargv;
extern char PROMPT_STR[];
extern char the_cold_msg[];
extern char the_warm_msg[];
extern struct encrypt_key this_key;

/*************************************************************
|               VECTORED WORDS
*/
extern char *(*FREAD_WA)();
extern char *(*READ_WA)();
extern char *(*Semi_colon)();

#ifdef VECTORED_IO
extern char *(*verbatim_vect)();
extern char *(*emit_vect)();
extern char *(*display_vect)();
extern char *(*center_vect)();
extern char *(*dot_quote_vect)();
extern char *(*type_vect)();
extern char *(*dot_vect)();
extern char *(*prints_vect)();
extern char *(*space_vect)();
extern char *(*spaces_vect)();
extern char *(*key_vect)();
extern char *(*query_terminal_vect)();
extern char *(*cr_vect)();
#endif
/*************************************************************
|
*/
extern int I;
extern union  cell_types  cell;

extern struct DictHeader **IP;                  /* Instrcution Pointer */
extern struct DictHeader *DP;                   /* Dictionary Pointer */
extern struct DictHeader *WA;                   /* Word Address Register */
extern struct DictHeader *pfa_list[];	/* Temp area to compile list of pfa's */
extern struct DictHeader *LIT_WA;               /* lit Word Address */
extern struct DictHeader *ZERO_BRAN_WA;
extern struct DictHeader *BRANCH_WA;
extern struct DictHeader *QUOTE_WA;
extern struct DictHeader *DOT_QUOTE_WA;
extern struct DictHeader *LOOP_WA;
extern struct DictHeader *DO_WA;
extern struct DictHeader *I_WA;
extern struct DictHeader *NAME0_WA;
extern struct DictHeader *ABORT_QUOTE_WA;
extern struct DictHeader *PLUS_LOOP_WA;
extern struct DictHeader *J_WA;
extern struct DictHeader *K_WA;
extern struct DictHeader *TICK_WA;
extern struct DictHeader *DO_COMPILE_WA;
extern struct DictHeader *DO_DOES_WA;
extern struct DictHeader *COMPILE_DO_DOES_WA;
extern struct DictHeader *DO_VOC_WA;
extern struct DictHeader *DO_DEF_WA;
extern struct DictHeader *DP0;
extern struct DictHeader *FORTH_WA;

extern struct NameField  *NAME0;

#endif
